
//////////////////////////////////////////////////////////////////////////

//Copyright @2003 Peng He,Information Science Insititute,XiDian University

//MyUsb example

/////////////////////////////////////////////////////////////////////////

//UsbDispatch.cpp:
//          Routines that contacked with the hardware Most of these 
//          routines response the application's I/O request

/////////////////////////////////////////////////////////////////////////

//Version history

//17-Apr-2003 creation

/////////////////////////////////////////////////////////////////////////




#include "UsbChip.h"
#include "Ioctl.h"


NTSTATUS UsbCreate(	IN PDEVICE_OBJECT   fdo,
					IN PIRP   Irp)
{
	PUSB_DEVICE_EXTENSION   dx = (PUSB_DEVICE_EXTENSION)fdo->DeviceExtension;
	PIO_STACK_LOCATION   IrpStack = IoGetCurrentIrpStackLocation(Irp);
	DebugPrint( "Create File is %T", &(IrpStack->FileObject->FileName));

	if( dx->IODisabled)
		return CompleteIrp( Irp, STATUS_DEVICE_NOT_CONNECTED, 0);

	InterlockedIncrement(&dx->OpenHandleCount);

	// Display USB host controller driver info
	USBD_VERSION_INFORMATION   vi;
	USBD_GetUSBDIVersion(&vi);
	DebugPrint("USBDI version %x Supported version %x", vi.USBDI_Version, vi.Supported_USB_Version);

	// Reset device and select configuration
	UsbResetDevice(dx);
	UsbSelectConfiguration(dx);


	return CompleteIrp(Irp,STATUS_SUCCESS,0);
}



NTSTATUS UsbClose(	IN PDEVICE_OBJECT fdo,
					IN PIRP Irp)
{
	PUSB_DEVICE_EXTENSION dx = (PUSB_DEVICE_EXTENSION)fdo->DeviceExtension;
	DebugPrintMsg("Close");

	InterlockedDecrement(&dx->OpenHandleCount);

	UsbDeselectConfiguration(dx);

	// Complete successfully
	return CompleteIrp(Irp,STATUS_SUCCESS,0);
}
 


NTSTATUS UsbRead(	IN PDEVICE_OBJECT   fdo,
					IN PIRP   Irp)
{
	PUSB_DEVICE_EXTENSION  dx=(PUSB_DEVICE_EXTENSION)fdo->DeviceExtension;
	if( dx->IODisabled)
		return CompleteIrp( Irp, STATUS_DEVICE_NOT_CONNECTED, 0);
	if (!LockDevice(dx))
		return CompleteIrp( Irp, STATUS_DELETE_PENDING, 0);

	PIO_STACK_LOCATION   IrpStack = IoGetCurrentIrpStackLocation(Irp);
	NTSTATUS   status = STATUS_SUCCESS;
	ULONG   BytesTxd = 0;

	// Get call parameters
	LONGLONG   FilePointer = IrpStack->Parameters.Read.ByteOffset.QuadPart;
	ULONG   ReadLen = IrpStack->Parameters.Read.Length;
	DebugPrint("Read %d bytes from file pointer %d",(int)ReadLen,(int)FilePointer);

	// Check file pointer
	if( FilePointer<0)
		status = STATUS_INVALID_PARAMETER;
	else
	{
		status = UsbReadData( dx, Irp->AssociatedIrp.SystemBuffer, ReadLen);
		BytesTxd = ReadLen;
	}
	DebugPrint("Read: %x %d bytes returned",status,(int)BytesTxd);

	if (status==STATUS_TIMEOUT)
		IoCancelIrp(Irp);	//ʱ,ȡIRP
	// Complete IRP
	CompleteIrp(Irp,status,BytesTxd);
	UnlockDevice(dx);
	return status;
}


NTSTATUS UsbWrite(	IN PDEVICE_OBJECT fdo,
					IN PIRP Irp)
{
	PUSB_DEVICE_EXTENSION   dx=(PUSB_DEVICE_EXTENSION)fdo->DeviceExtension;
	if( dx->IODisabled)
		return CompleteIrp( Irp, STATUS_DEVICE_NOT_CONNECTED, 0);
	if (!LockDevice(dx))
		return CompleteIrp( Irp, STATUS_DELETE_PENDING, 0);

	PIO_STACK_LOCATION   IrpStack = IoGetCurrentIrpStackLocation(Irp);
	NTSTATUS   status = STATUS_SUCCESS;
	ULONG BytesTxd = 0;

	// Get call parameters
	LONGLONG   FilePointer = IrpStack->Parameters.Write.ByteOffset.QuadPart;
	ULONG   WriteLen = IrpStack->Parameters.Write.Length;
	DebugPrint("Write %d bytes from file pointer %d",(int)WriteLen,(int)FilePointer);

	if( FilePointer<0 || WriteLen<1)
		status = STATUS_INVALID_PARAMETER;
	else
	{
		status = UsbWriteData( dx, Irp->AssociatedIrp.SystemBuffer, WriteLen);
		BytesTxd = WriteLen;


	}

	DebugPrint("Write: %d bytes written",(int)BytesTxd);

	if (status==STATUS_TIMEOUT)
		IoCancelIrp(Irp);	//ʱ,ȡIRP

	// Complete IRP
	CompleteIrp(Irp,status,BytesTxd);
	UnlockDevice(dx);
	return status;
}


NTSTATUS UsbDeviceControl(	IN PDEVICE_OBJECT fdo,
							IN PIRP Irp)
{
	PUSB_DEVICE_EXTENSION   dx = (PUSB_DEVICE_EXTENSION)fdo->DeviceExtension;
	if( dx->IODisabled)
		return CompleteIrp( Irp, STATUS_DEVICE_NOT_CONNECTED, 0);
	if (!LockDevice(dx))
		return CompleteIrp( Irp, STATUS_DELETE_PENDING, 0);

	PIO_STACK_LOCATION   IrpStack = IoGetCurrentIrpStackLocation(Irp);
	NTSTATUS   status = STATUS_SUCCESS;
	ULONG   BytesTxd = 0;

	ULONG   ControlCode = IrpStack->Parameters.DeviceIoControl.IoControlCode;
	ULONG   InputLength = IrpStack->Parameters.DeviceIoControl.InputBufferLength;
	ULONG   OutputLength = IrpStack->Parameters.DeviceIoControl.OutputBufferLength;

	DebugPrint("DeviceIoControl: Control code %x InputLength %d OutputLength %d",
				ControlCode, InputLength, OutputLength);

	switch( ControlCode)
	{

	///////	Get device descriptor
	case IOCTL_USB_GET_DEVICE_DESCRIPTOR:
	{
		PUSB_DEVICE_DESCRIPTOR deviceDescriptor = NULL;
		ULONG size;
		NTSTATUS status = UsbGetDeviceDescriptor( dx, deviceDescriptor, size);
		if( NT_SUCCESS(status))
		{
			BytesTxd = size;
			if( BytesTxd>OutputLength)
				BytesTxd = OutputLength;
			RtlCopyMemory(Irp->AssociatedIrp.SystemBuffer,deviceDescriptor,BytesTxd);
		}
		FreeIfAllocated(deviceDescriptor);
		break;
	}

	///////	Get configuration and other descriptors
	case IOCTL_USB_GET_CONFIGURATION_DESCRIPTORS:
	{
		PUSB_CONFIGURATION_DESCRIPTOR Descriptors = NULL;
		ULONG size;
		NTSTATUS status = UsbGetConfigurationDescriptors( dx, Descriptors, 0, size);
		if( NT_SUCCESS(status))
		{
			BytesTxd = size;
			if( BytesTxd>OutputLength)
				BytesTxd = OutputLength;
			RtlCopyMemory(Irp->AssociatedIrp.SystemBuffer,Descriptors,BytesTxd);
		}
		FreeIfAllocated(Descriptors);
		break;
	}


	///////	Get Statuses
	case IOCTL_USB_GET_STATUSES:
	{
		UCHAR Statuses[6];
		ULONG size = sizeof(Statuses);
		NTSTATUS status = UsbGetStatuses( dx, Statuses, size);
		if( NT_SUCCESS(status))
		{
			BytesTxd = size;
			if( BytesTxd>OutputLength)
				BytesTxd = OutputLength;
			RtlCopyMemory(Irp->AssociatedIrp.SystemBuffer,Statuses,BytesTxd);
		}
		break;
	}


	///////	Invalid request
	default:
		status = STATUS_INVALID_DEVICE_REQUEST;
	}

	DebugPrint("DeviceIoControl: %d bytes written",(int)BytesTxd);

	// Complete IRP
	CompleteIrp(Irp,status,BytesTxd);
	UnlockDevice(dx);
	return status;
}


NTSTATUS UsbSystemControl(	IN PDEVICE_OBJECT   fdo,
							IN PIRP   Irp)
{
	PUSB_DEVICE_EXTENSION dx = (PUSB_DEVICE_EXTENSION)fdo->DeviceExtension;
	if( dx->IODisabled)
		return CompleteIrp( Irp, STATUS_DEVICE_NOT_CONNECTED, 0);
	if (!LockDevice(dx))
		return CompleteIrp( Irp, STATUS_DELETE_PENDING, 0);

	DebugPrintMsg("SystemControl");

	// Just pass to lower driver
	IoSkipCurrentIrpStackLocation(Irp);
	NTSTATUS status = IoCallDriver( dx->NextStackDevice, Irp);
	UnlockDevice(dx);
	return status;
}


NTSTATUS CompleteIrp( IN PIRP Irp, IN NTSTATUS status, IN ULONG info)
{
	Irp->IoStatus.Status = status;
	Irp->IoStatus.Information = info;
	IoCompleteRequest(Irp,IO_NO_INCREMENT);
	return status;
}
