#define DATA_IN_PIPE             1    // USB9602 EndPoint1 for tranmit
#define DATA_OUT_PIPE            2    // USB9602 EndPoint2 for recieve
#define MAXIMUM_TRANSFER_SIZE   1024	//Maximun transfer size in one transaction



#ifdef __cplusplus
extern "C"
{
#endif
#include "wdm.h"
#include "usb100.h"
#include "usbdi.h"
#include "usbdlib.h"
#ifdef __cplusplus
}
#endif


#include "DebugPrint.h"

#include "Guids.h"


typedef struct _USB_DEVICE_EXTENSION
{
	PDEVICE_OBJECT	fdo;
	PDEVICE_OBJECT	pdo;
	PDEVICE_OBJECT	NextStackDevice;
	UNICODE_STRING	ifSymLinkName;

	bool GotResources;	// Not stopped
	bool Paused;		// Stop or remove pending
	bool IODisabled;	// Paused or stopped
//	bool InterruptsEnabled;

	LONG OpenHandleCount;	// Count of open handles

	LONG UsageCount;		// Pending I/O Count
	bool Stopping;			// In process of stopping
	KEVENT StoppingEvent;	// Set when all pending I/O complete

	USBD_CONFIGURATION_HANDLE   UsbConfigurationHandle;	// Selected Configuration handle
	USBD_PIPE_HANDLE   UsbPipeHandle;			// Handle to input interrupt pipe
	USBD_PIPE_HANDLE   UsbDataInHandle;
	USBD_PIPE_HANDLE   UsbDataOutHandle;
	ULONG UsbTimeout;	// Read timeout in seconds

	// Resource allocations
	bool GotPortOrMemory;
	bool PortInIOSpace;
	bool PortNeedsMapping;
	PUCHAR PortBase;
	PHYSICAL_ADDRESS PortStartAddress;
	ULONG PortLength;

	bool GotInterrupt;
	ULONG Vector;
	KIRQL Irql;
	KINTERRUPT_MODE Mode;
	KAFFINITY Affinity;
	PKINTERRUPT InterruptObject;

} USB_DEVICE_EXTENSION, *PUSB_DEVICE_EXTENSION;


VOID UsbUnload(IN PDRIVER_OBJECT DriverObject);

NTSTATUS UsbPower(	IN PDEVICE_OBJECT fdo,
					IN PIRP Irp);

NTSTATUS UsbPnp(	IN PDEVICE_OBJECT fdo,
					IN PIRP Irp);

NTSTATUS UsbAddDevice(	IN PDRIVER_OBJECT DriverObject,
						IN PDEVICE_OBJECT pdo);

NTSTATUS UsbCreate( IN PDEVICE_OBJECT fdo,
					IN PIRP Irp);

NTSTATUS UsbClose(	IN PDEVICE_OBJECT fdo,
					IN PIRP Irp);
 
NTSTATUS UsbWrite(	IN PDEVICE_OBJECT fdo,
					IN PIRP Irp);

NTSTATUS UsbRead(	IN PDEVICE_OBJECT fdo,
					IN PIRP Irp);

NTSTATUS UsbDeviceControl(	IN PDEVICE_OBJECT fdo,
							IN PIRP Irp);

NTSTATUS UsbSystemControl(	IN PDEVICE_OBJECT fdo,
							IN PIRP Irp);


bool LockDevice( IN PUSB_DEVICE_EXTENSION dx);
void UnlockDevice( IN PUSB_DEVICE_EXTENSION dx);

NTSTATUS StartDevice(	IN PUSB_DEVICE_EXTENSION dx,
						IN PCM_RESOURCE_LIST AllocatedResourcesTranslated);

VOID StopDevice( IN PUSB_DEVICE_EXTENSION dx);

NTSTATUS ForwardIrpAndWait( IN PDEVICE_OBJECT fdo, IN PIRP Irp);


_declspec(dllimport) NTSTATUS UsbResetDevice( IN PUSB_DEVICE_EXTENSION dx);

NTSTATUS UsbGetUsbInfo( IN PUSB_DEVICE_EXTENSION dx);

NTSTATUS UsbGetDeviceDescriptor( IN PUSB_DEVICE_EXTENSION dx,
								 OUT PUSB_DEVICE_DESCRIPTOR& deviceDescriptor,
								 OUT ULONG& Size);

NTSTATUS UsbGetConfigurationDescriptors( IN PUSB_DEVICE_EXTENSION dx, 
										 OUT PUSB_CONFIGURATION_DESCRIPTOR& descriptors,
										 IN UCHAR ConfigIndex,
										 OUT ULONG& DescriptorsSize);

NTSTATUS UsbGetSpecifiedDescriptor( IN PUSB_DEVICE_EXTENSION dx,
									OUT PVOID& Descriptor,
									IN UCHAR DescriptorType,
									IN OUT ULONG& Size);

NTSTATUS UsbSelectConfiguration( IN PUSB_DEVICE_EXTENSION dx);

NTSTATUS UsbDeselectConfiguration( IN PUSB_DEVICE_EXTENSION dx);

NTSTATUS UsbReadData( IN PUSB_DEVICE_EXTENSION dx, IN PVOID Buffer, ULONG& BufferSize);


NTSTATUS UsbWriteData( IN PUSB_DEVICE_EXTENSION dx, IN PVOID UserBuffer, ULONG& UserBufferSize);


NTSTATUS UsbGetStatuses( IN PUSB_DEVICE_EXTENSION dx,
						 OUT PUCHAR Statuses,
						 IN ULONG Size);

NTSTATUS UsbGetIdleRate( IN PUSB_DEVICE_EXTENSION dx);

#define FreeIfAllocated(x) if( (x)!=NULL) { ExFreePool(x); (x) = NULL; }


NTSTATUS CompleteIrp( IN PIRP Irp, IN NTSTATUS status, IN ULONG info);